import { FormGroup, FormControl, FormBuilder } from '@angular/forms';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';
import { MetadataStorage, Validator, getFromContainer, ValidationTypes, validateSync } from 'class-validator';
import { classToClass, plainToClass } from 'class-transformer';
import 'reflect-metadata';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * @template TModel
 */
class DynamicFormGroup extends FormGroup {
    /**
     * @param {?} factoryModel
     * @param {?} fields
     */
    constructor(factoryModel, fields) {
        super({});
        this.factoryModel = factoryModel;
        this.fields = fields;
        this.otherErrors = [];
        this.customValidateErrors = new BehaviorSubject({});
        this.fields = this.onlyFields(this.fields);
    }
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?} fields
     * @param {?=} validatorOptions
     * @return {?}
     */
    static getClassValidators(factoryModel, fields, validatorOptions) {
        const /** @type {?} */ validationMetadatas = getFromContainer(MetadataStorage).getTargetValidationMetadatas(factoryModel, '', validatorOptions && validatorOptions.groups ? validatorOptions.groups : undefined);
        const /** @type {?} */ formGroupFields = {};
        const /** @type {?} */ validator = new Validator();
        Object.keys(fields).filter(key => key.indexOf('__') !== 0).forEach(key => {
            let /** @type {?} */ formGroupField = formGroupFields[key];
            if (formGroupField === undefined) {
                formGroupField = Array.isArray(fields[key]) ? fields[key] : [];
            }
            if (!Array.isArray(fields[key])) {
                if (fields[key] instanceof DynamicFormGroup) {
                    fields[key].object = fields[key].fields;
                    formGroupField.push(fields[key]);
                }
                else {
                    formGroupField.push(fields[key]);
                }
            }
            validationMetadatas.forEach(validationMetadata => {
                if (validationMetadata.propertyName === key) {
                    switch (validationMetadata.type) {
                        /*
                                                case ValidationTypes.IS_NOT_EMPTY: {
                                                    formGroupField.push(Validators.required);
                                                    break;
                                                }
                                                case ValidationTypes.IS_EMAIL: {
                                                    formGroupField.push(Validators.required);
                                                    break;
                                                }*/
                        default: {
                            for (const /** @type {?} */ typeKey in ValidationTypes) {
                                if (ValidationTypes.hasOwnProperty(typeKey)) {
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] === undefined &&
                                        validationMetadata.type === 'nestedValidation' &&
                                        typeKey === 'NESTED_VALIDATION') {
                                        const /** @type {?} */ nestedValidate = function (c) {
                                            const /** @type {?} */ isValid = (c.parent && c.parent.value) ?
                                                validateSync(c.value, validatorOptions).length === 0 : true;
                                            return isValid ? null : {
                                                nestedValidate: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(nestedValidate);
                                    }
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] === undefined &&
                                        validationMetadata.type === 'customValidation' &&
                                        typeKey === 'CUSTOM_VALIDATION') {
                                        const /** @type {?} */ customValidation = function (c) {
                                            // todo: refactor
                                            const /** @type {?} */ object = (c.parent instanceof DynamicFormGroup) ?
                                                (/** @type {?} */ (c.parent)).object :
                                                (c.parent ? c.parent.value : {});
                                            if (object) {
                                                object[key] = c.value;
                                            }
                                            const /** @type {?} */ validateErrors = (c.parent && c.parent.value) ?
                                                validateSync(object, validatorOptions) : [];
                                            const /** @type {?} */ isValid = validateErrors.filter((error) => {
                                                if (error.children.length &&
                                                    error.children.filter(children => children.property === key)) {
                                                    return true;
                                                }
                                                return true;
                                            }).length === 0;
                                            return isValid ? null : {
                                                customValidation: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(customValidation);
                                    }
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] !== undefined) {
                                        const /** @type {?} */ customValidate = function (c) {
                                            if (!c) {
                                                return null;
                                            }
                                            const /** @type {?} */ isValid = c.parent && c.parent.value ? validator[validationMetadata.type](c.value) : true;
                                            return isValid ? null : {
                                                customValidate: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(customValidate);
                                    }
                                }
                            }
                            break;
                        }
                    }
                }
            });
            if (formGroupField[0] instanceof DynamicFormGroup) {
                formGroupFields[key] = formGroupField[0];
            }
            else {
                formGroupFields[key] = [
                    formGroupField[0],
                    formGroupField.filter((item, index) => index !== 0)
                ];
            }
        });
        return formGroupFields;
    }
    /**
     * @param {?} fields
     * @return {?}
     */
    onlyFields(fields) {
        const /** @type {?} */ newFields = {};
        if (fields !== undefined) {
            Object.keys(fields).forEach(key => {
                if (fields[key] instanceof DynamicFormGroup) {
                    newFields[key] = this.onlyFields((/** @type {?} */ (fields[key])).fields);
                }
                else {
                    newFields[key] = fields[key];
                }
            });
        }
        return newFields;
    }
    /**
     * @return {?}
     */
    get object() {
        const /** @type {?} */ object = this._object ? this.classToClass(this._object) : new this.factoryModel();
        if (object !== undefined) {
            Object.keys(this.controls).forEach(key => {
                if (this.controls[key] instanceof DynamicFormGroup) {
                    object[key] = (/** @type {?} */ (this.controls[key])).object;
                }
                else {
                    object[key] = this.controls[key].value;
                }
            });
        }
        return this.plainToClass(this.factoryModel, object);
    }
    /**
     * @param {?} object
     * @return {?}
     */
    set object(object) {
        if (object instanceof this.factoryModel) {
            this._object = this.classToClass(object);
        }
        else {
            this._object = this.plainToClass(this.factoryModel, /** @type {?} */ (object));
        }
        Object.keys(this.controls).forEach(key => {
            if (this.controls[key] instanceof DynamicFormGroup) {
                (/** @type {?} */ (this.controls[key])).object = this._object ? this._object[key] : {};
            }
            else {
                this.controls[key].setValue((this._object && this._object[key]) ? this._object[key] : undefined);
            }
        });
    }
    /**
     * @param {?} errors
     * @return {?}
     */
    transformValidationErrors(errors) {
        const /** @type {?} */ customErrors = {};
        errors.forEach((error) => {
            if (error.constraints !== undefined) {
                Object.keys(error.constraints).forEach((key) => {
                    if (!customErrors[error.property]) {
                        customErrors[error.property] = [];
                    }
                    if (customErrors[error.property].indexOf(error.constraints[key]) === -1) {
                        customErrors[error.property].push(error.constraints[key]);
                    }
                });
            }
            if (error.children !== undefined && error.children.length) {
                customErrors[error.property] = this.transformValidationErrors(error.children);
            }
        });
        return customErrors;
    }
    /**
     * @template TClassModel
     * @param {?} object
     * @return {?}
     */
    classToClass(object) {
        return classToClass(object, { ignoreDecorators: true });
    }
    /**
     * @template TClassModel, Object
     * @param {?} cls
     * @param {?} plain
     * @return {?}
     */
    plainToClass(cls, plain) {
        return plainToClass(cls, plain, { ignoreDecorators: true });
    }
    /**
     * @param {?=} otherErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    validate(otherErrors, validatorOptions) {
        if (otherErrors === undefined) {
            otherErrors = this.otherErrors;
        }
        const /** @type {?} */ errors = validateSync(this.object, validatorOptions);
        this.customValidateErrors.next(this.transformValidationErrors([...errors, ...otherErrors]));
    }
    /**
     * @return {?}
     */
    validateAllFormFields() {
        Object.keys(this.controls).forEach(field => {
            const /** @type {?} */ control = this.get(field);
            if (control instanceof FormControl) {
                control.markAsTouched({ onlySelf: true });
            }
            else if (control instanceof DynamicFormGroup) {
                control.validateAllFormFields();
            }
        });
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class DynamicFormBuilder extends FormBuilder {
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} controlsConfig
     * @param {?=} extra
     * @return {?}
     */
    group(factoryModel, controlsConfig, extra) {
        if (extra !== undefined && extra.customValidatorOptions === undefined) {
            extra.customValidatorOptions = { validationError: { target: false } };
        }
        let /** @type {?} */ newControlsConfig;
        if (controlsConfig !== undefined) {
            newControlsConfig = controlsConfig;
        }
        // experimental
        if (controlsConfig === undefined) {
            newControlsConfig = new factoryModel();
            Object.keys(newControlsConfig).forEach(key => {
                if (newControlsConfig[key] &&
                    newControlsConfig[key].constructor &&
                    typeof newControlsConfig[key] === 'object' &&
                    (newControlsConfig[key].length === undefined ||
                        (newControlsConfig[key].length !== undefined &&
                            Object.keys(newControlsConfig[key].length).length === newControlsConfig[key].length))) {
                    newControlsConfig[key] = this.group(newControlsConfig[key].constructor, undefined, extra);
                }
            });
        }
        const /** @type {?} */ dynamicFormGroup = new DynamicFormGroup(factoryModel, newControlsConfig);
        const /** @type {?} */ formGroup = super.group(DynamicFormGroup.getClassValidators(factoryModel, newControlsConfig, (extra && extra.customValidatorOptions) ? extra.customValidatorOptions : undefined), extra);
        Object.keys(formGroup.controls).forEach(key => {
            dynamicFormGroup.addControl(key, formGroup.controls[key]);
        });
        dynamicFormGroup.valueChanges.subscribe(data => {
            dynamicFormGroup.validate(undefined, (extra && extra.customValidatorOptions) ? extra.customValidatorOptions : undefined);
        });
        return dynamicFormGroup;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
const NgxDynamicFormBuilderShareds = [DynamicFormBuilder, DynamicFormGroup];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * Generated bundle index. Do not edit.
 */

export { DynamicFormBuilder, DynamicFormGroup, NgxDynamicFormBuilderShareds };
//# sourceMappingURL=ngx-dynamic-form-builder.js.map
